local function DEBUG(msg) 
   DEFAULT_CHAT_FRAME:AddMessage("[SROD] " .. msg);
end

local LINEFACTOR = 32/30; -- Multiplying factor for texture coordinates
local LINEFACTOR_2 = LINEFACTOR / 2; -- Half o that

local LOG2 = math.log(2);

local SCALE_SNAPTO = 0.01; -- Scale bits
local ANGLE_SNAPTO = 1.5; -- Degrees
local OFFSET_SNAPTO = 0.02; -- Offset bits.

local MIN_SCALE = 0.04; -- Smallest scale 4%

local STANDARD_RADIUS = 100;
local THIN_LINE_WIDTH = 1;
local THICK_LINE_WIDTH = 10;

local function DoSnapTo(value, snapto)
   return snapto * math.floor(value / snapto + 0.5);
end

local function GetEffectiveLength(anchor)
   local al,as = anchor.length, anchor.scale;
   if (al <= 1) then
      return al / as;
   end
   return (1 / as) + (al - 1);
end

local function SetEffectiveLength(anchor, el)
   local as = 1 / anchor.scale;
   if (el > as) then
      anchor.length = (el-as) + 1;
   else
      anchor.length = el / as;
   end
end

-- We'll be calling these a lot, so grab local versions (Yes, some of them
-- are defined in compat.lua but I can't count on that can I?)
local rad, deg = math.rad, math.deg;
local sin, cos, atan2 = math.sin, math.cos, math.atan2;
local sqrt,abs = math.sqrt, math.abs;

local function SinCosDeg(deg)
   local r = rad(deg);
   return sin(r), cos(r);
end

-- T        - Texture
-- C        - Canvas Frame (for anchoring)
-- sx,sy    - Coordinate of start of line
-- ex,ey    - Coordinate of end of line
-- w        - Width of line
-- relPoint - Relative point on canvas to interpret coords (Default BOTTOMLEFT)
local function DrawLine(T, C, sx, sy, ex, ey, w, relPoint)
   if (not relPoint) then relPoint = "BOTTOMLEFT"; end

   -- Determine dimensions and center point of line
   local dx,dy = ex - sx, ey - sy;
   local cx,cy = (sx + ex) / 2, (sy + ey) / 2;

   -- Normalize direction if necessary
   if (dx < 0) then
      dx,dy = -dx,-dy;
   end

   -- Calculate actual length of line
   local l = sqrt((dx * dx) + (dy * dy));

   -- Quick escape if it's zero length
   if (l == 0) then
      T:SetTexCoord(0,0,0,0,0,0,0,0);
      T:SetPoint("BOTTOMLEFT", C, relPoint, cx,cy);
      T:SetPoint("TOPRIGHT",   C, relPoint, cx,cy);
      return;
   end

   -- Sin and Cosine of rotation, and combination (for later)
   local s,c = -dy / l, dx / l;
   local sc = s * c;

   -- Calculate bounding box size and texture coordinates
   local Bwid, Bhgt, BLx, BLy, TLx, TLy, TRx, TRy, BRx, BRy;
   if (dy >= 0) then
      Bwid = ((l * c) - (w * s)) * LINEFACTOR_2;
      Bhgt = ((w * c) - (l * s)) * LINEFACTOR_2;
      BLx, BLy, BRy = (w / l) * sc, s * s, (l / w) * sc;
      BRx, TLx, TLy, TRx = 1 - BLy, BLy, 1 - BRy, 1 - BLx; 
      TRy = BRx;
   else
      Bwid = ((l * c) + (w * s)) * LINEFACTOR_2;
      Bhgt = ((w * c) + (l * s)) * LINEFACTOR_2;
      BLx, BLy, BRx = s * s, -(l / w) * sc, 1 + (w / l) * sc;
      BRy, TLx, TLy, TRy = BLx, 1 - BRx, 1 - BLx, 1 - BLy;
      TRx = TLy;
   end

   -- Set texture coordinates and anchors
   T:SetTexCoord(TLx, TLy, BLx, BLy, TRx, TRy, BRx, BRy);
   T:SetPoint("BOTTOMLEFT", C, relPoint, cx - Bwid, cy - Bhgt);
   T:SetPoint("TOPRIGHT",   C, relPoint, cx + Bwid, cy + Bhgt);
end

local topTextures = {};
local midTextures = {};
local botTextures = {};

function StatRingsOptionsDrag_OnLoad()
   this:SetAlpha(1.0);

   local t=0;
   while true do
      t = t + 1;
      local T = getglobal("StatRingsOptionsDragFrameO" .. t);
      if (not T) then break; end
      T:ClearAllPoints();
      topTextures[t] = T;
   end
   t = 0;
   while true do
      t = t + 1;
      local T = getglobal("StatRingsOptionsDragFrameA" .. t);
      if (not T) then break; end
      T:ClearAllPoints();
      if (t > 4) then
	 T:SetVertexColor(1,0.5,0,0.9);
      else
	 T:SetVertexColor(1,0.5,0,0.6);
      end
      midTextures[t] = T;
   end
   t = 0;
   while true do
      t = t + 1;
      local T = getglobal("StatRingsOptionsDragFrameB" .. t);
      if (not T) then break; end
      T:ClearAllPoints();
      T:SetVertexColor(0.6,0.6,1,0.7);
      --T:SetVertexColor(1,1,1,0.5);
      botTextures[t] = T;
   end

   this.infoFrame = StatRingsOptionsDragFrameInfo;
   this.textFrames = {};
   for i=1,5 do
      this.textFrames[i] = getglobal("StatRingsOptionsDragFrameInfoT" .. i);
   end

   this.drag = StatRingsOptionsDragFrameDrag;
   this.drag:SetAlpha(0.5);

   this.frameName = this:GetName();
end

local ANCHOR_FIELDS = {
   "length", "angle", "spin", "scale" , "relSpin"
};

function StatRingsOptionsDrag_Start(label, parent,
				    rootX, rootY, rootScale, rootSpin,
				    anchor, callback)
   local this = StatRingsOptionsDragFrame;

   this.rootX, this.rootY = rootX, rootY; -- Root position (relative to parent)
   this.rootScale = rootScale; -- Effective root scale (relative to parent)
   this.rootSpin = rootSpin;    -- Effective root spin
   
   this:SetParent(parent);
   this:SetFrameStrata("FULLSCREEN") -- Force it back
   this.infoFrame:SetFrameStrata("FULLSCREEN_DIALOG") -- Force it back
   this.relativeTo = parent;
   this.relativeToName = parent:GetName();

   this.updateCallback = callback;

   this.anchor = anchor;
   if (not this.anchorRef) then
      this.anchorRef = {};
   end

   -- Save for comparison
   for _,k in ipairs(ANCHOR_FIELDS) do
      this.anchorRef[k] = anchor[k];
   end
   
   this.textFrames[1]:SetText(label);

   StatRingsOptionsDrag_UpdateFrame(this);

   this:Show();
end

local arrS,arrC = SinCosDeg(30);
local arrL = 0.2;
-- 1:Origin  2-5:Corners 6-9:Edge Centers
local BOX_X = {0,   -1, 1,1,-1,   0,  1,  0, -1};
local BOX_Y = {0,   -1,-1,1, 1,   1,  0, -1,  0};
local ROTBOX_OTH = {6, 3, 4, 5, 2}
local ROTBOX_CNT = 6;
local GRIDBOX_OTH = {0, 3, 4, 5, 2, 8, 9}
local GRIDBOX_CNT = 9;
-- Arrow head coordinates
local ARROW_X = { -arrS*arrL, 0, arrS*arrL }
local ARROW_Y = { 1-arrC*arrL, 1, 1-arrC*arrL }
local ARROW_OTH = {2,3}
local ARROW_CNT = 3;

local xpts,ypts = {},{};

local function drawArrowBox(name, s, c, size, xo, yo, width, textureSet, tx)
   tx = tx or 0;
   
   -- Calculate child rotated box
   for i=1,ROTBOX_CNT do
      local x,y = BOX_X[i] * size, BOX_Y[i] * size;
      xpts[i] = x * c + y * s + xo;
      ypts[i] = y * c - x * s + yo;
   end
   -- Draw outer box and turn line
   for i,j in ipairs(ROTBOX_OTH) do
      tx = tx + 1;
      local T = textureSet[tx];
      DrawLine(T, name, xpts[i], ypts[i], xpts[j], ypts[j], width, 
	       "CENTER");
      T:Show();
   end
   -- Calculate arrow head
   for i=1,ARROW_CNT do
      local x,y = ARROW_X[i] * size, ARROW_Y[i] * size;
      xpts[i] = x * c + y * s + xo;
      ypts[i] = y * c - x * s + yo;
   end
   -- Draw arrow head
   for i,j in ipairs(ARROW_OTH) do
      tx = tx + 1;
      local T = textureSet[tx];
      DrawLine(T, name, xpts[i], ypts[i], xpts[j], ypts[j], width, 
	       "CENTER");
      T:Show();
   end
   return tx;
end

local function drawGridBox(name, size, xo, yo, width, textureSet, tx)
   tx = tx or 0;
   for i=2,GRIDBOX_CNT do
      local x,y = BOX_X[i] * size, BOX_Y[i] * size;
      xpts[i] = x + xo;
      ypts[i] = y + yo;
   end
   for i,j in ipairs(GRIDBOX_OTH) do
      if (i > 1) then
	 tx = tx + 1;     
	 local T = textureSet[tx];
	 DrawLine(T, name, xpts[i], ypts[i], xpts[j], ypts[j], width, 
		  "CENTER");
	 T:Show();
      end
   end
   return tx;
end

function StatRingsOptionsDrag_UpdateFrame(this)
   local name = this.frameName or this:GetName();
   local uiScale = UIParent:GetScale();

   local anchor = this.anchor;
   local anchorScale = anchor.scale;

   -- Scale this against the parent by the effective scale
   this:SetScale(anchorScale * this.rootScale);

   -- Work out the relative scale of this frame against its parent
   local effScale = this:GetEffectiveScale();
   local scaleRatio = this.relativeTo:GetEffectiveScale() / effScale;
   -- Normalize against uiScale now
   effScale = effScale / uiScale;

   this.infoFrame:SetScale(1 / effScale);

   -- Calculate actual widths to achieve these pixel widths
   local thinWidth = THIN_LINE_WIDTH / effScale;
   local thickWidth = THICK_LINE_WIDTH / effScale;

   local spin = anchor.spin;
   if (anchor.relSpin) then
      spin = spin + anchor.angle;
   end

   local s,c = SinCosDeg(spin);
   local ofsS,ofsC = SinCosDeg(anchor.angle);
   local maxRadius = (abs(c) + abs(s)) * (STANDARD_RADIUS + thinWidth / 2);

   local ro = GetEffectiveLength(anchor) * STANDARD_RADIUS;
   
   local roX = -ro * ofsS;
   local roY = -ro * ofsC;

   this:ClearAllPoints();
   this:SetPoint("CENTER", this.relativeTo, "CENTER",
		 this.rootX * scaleRatio - roX,
		 this.rootY * scaleRatio - roY);
   this:SetWidth(maxRadius * 2);
   this:SetHeight(maxRadius * 2);

   this.infoFrame:ClearAllPoints();
   this.infoFrame:SetPoint("TOPLEFT", name, "CENTER",
			   STANDARD_RADIUS * effScale + 10,
			   STANDARD_RADIUS * effScale);
   
   drawArrowBox(name, s, c, STANDARD_RADIUS, 0, 0, thinWidth, topTextures);

   local tx = 0;
   -- Calculate inner drag box
   local thickRad = STANDARD_RADIUS - thickWidth / 2;
   for i=2,5 do
      local x,y = BOX_X[i] * thickRad, BOX_Y[i] * thickRad;
      xpts[i] = x * c + y * s;
      ypts[i] = y * c - x * s;
   end
   for i=2,5 do
      local j = ROTBOX_OTH[i];
      tx = tx + 1;     
      local T = midTextures[tx];
      DrawLine(T, name, xpts[i], ypts[i], xpts[j], ypts[j], thickWidth, 
	       "CENTER");
      T:Show();
   end
   tx = drawGridBox(name, STANDARD_RADIUS, 0, 0, thinWidth, midTextures, tx);

   -- Calculate rotated origin box
   s,c = SinCosDeg(this.rootSpin);
   tx = drawArrowBox(name, s, c, STANDARD_RADIUS / anchorScale, 
		     roX, roY, thinWidth, botTextures);
   tx = drawGridBox(name, STANDARD_RADIUS / anchorScale, 
		     roX, roY, thinWidth, botTextures, tx);

   this.textFrames[2]:SetText(format("scale: %.2f", anchor.scale));
   this.textFrames[3]:SetText(format("angle: %.1f", anchor.angle));
   this.textFrames[4]:SetText(format("length: %.2f", anchor.length));
   local rsText = "";
   if (anchor.relSpin) then
      rsText = " (relative)";
   end
   this.textFrames[5]:SetText(format("spin: %-.1f%s", anchor.spin, rsText));
   
   local hasChanged = nil;
   local anchorRef = this.anchorRef;
   -- Save for comparison
   for _,k in ipairs(ANCHOR_FIELDS) do
      if (anchorRef[k] ~= anchor[k]) then
	 hasChanged = true;
	 break;
      end
   end

   if (hasChanged) then
      for _,k in ipairs(ANCHOR_FIELDS) do
	 anchorRef[k] = anchor[k];
      end
      if (this.updateCallback) then
	 this.updateCallback(anchorRef);
      end
   end

--   -- Draw drag line from origin
--   local T = botTextures[5];
--   DrawLine(T, name, roX, roY, 0, 0, thickWidth/2, "CENTER");
--   T:Show();
end

function StatRingsOptionsDrag_OnMouseDown()
   if ((arg1 ~= "LeftButton") and (arg1 ~= "RightButton")) then
      return;
   end

   local drag = this.drag;
   local anchor = this.anchor;

   -- See if dragging is active already
   if (drag.mode) then
      -- We already know about this one
      if (drag[arg1]) then
	 return;
      end
      drag[arg1] = 1;
      drag.buttons = drag.buttons + 1;
      return;
   end

   local ox,oy = GetCursorPosition();
   local effScale = this:GetEffectiveScale();
   local uiScale = UIParent:GetScale();
   local x,y = ox/effScale, oy/effScale;
   local midX, midY = this:GetCenter();
   x,y = x - midX, y - midY;

   local thickWidth = THICK_LINE_WIDTH * uiScale / effScale;
   local thinWidth = THIN_LINE_WIDTH * uiScale / effScale;

   local spin = anchor.spin;
   if (anchor.relSpin) then
      spin = spin + anchor.angle;
   end
   local s,c = SinCosDeg(spin);
   
   local mout = abs(x*c - y*s);
   local macross = abs(y*c + x*s);
   if (macross > mout) then
      macross,mout = mout,macross;
   end
   
   if (mout > (STANDARD_RADIUS+thinWidth)) then
      -- Click outside of the box for sure
      return;
   end

   local mode;

   if (mout < (STANDARD_RADIUS - thickWidth)) then
      mode = "move";
   else
      if (macross > 66) then
	 mode = "scaleRotate";
      elseif (arg1 == "LeftButton") then
	 mode = "rotate";
      else
	 mode = "scale";
      end
   end
   
   drag.snapTo = 1;
   local needUpdate = nil;

   if (mode == "move") then
      -- Handle MOVEMENT only for now
      -- We need the screen coordinate scale factor
      local ofsS,ofsC = SinCosDeg(anchor.angle);
      -- Root offset relative to frame
      local ro = GetEffectiveLength(anchor) * STANDARD_RADIUS;
      ro = ro * effScale;
      -- Root offset relative to window
      drag.startx, drag.starty = ox - ro * ofsS,oy - ro * ofsC;
      drag.ofsScale = effScale;
   elseif ((mode == "scaleRotate") 
	   or (mode == "scale") or (mode == "rotate")) then
      -- Handle SCALING and ROTATION
      local centerX,centerY = midX*effScale, midY*effScale;
      drag.startx, drag.starty = centerX, centerY;
      local dx,dy = (ox - centerX), (oy - centerY);
      local curLen = sqrt(dx*dx + dy*dy);
      drag.baseLen = curLen / anchor.scale;
      local curAngle = 0;
      if (curLen > 0) then
	 curAngle = deg(atan2(dx,dy));
      end
      
      if (mode == "scaleRotate" or mode == "scale") then
	 drag.doScale = 1;
      else
	 drag.doScale = nil;
      end

      if (mode == "scaleRotate" or mode == "rotate") then
	 drag.doAngle = 1;

	 if (IsAltKeyDown()) then
	    needUpdate = true;
	    if (anchor.relSpin) then
	       anchor.spin = math.mod(anchor.spin + anchor.angle, 360);
	       if (anchor.spin < 0) then
		  anchor.spin = anchor.spin + 360;
	       end
	       anchor.relSpin = nil;
	    else
	       anchor.spin = math.mod(anchor.spin - anchor.angle, 360);
	       if (anchor.spin < 0) then
		  anchor.spin = anchor.spin + 360;
	       end
	       anchor.relSpin = true;
	    end
	 end
      else
	 drag.doAngle = nil;
      end

      drag.ofsAngle = anchor.spin - curAngle;

      mode = "scaleRotate";
   else
      mode = nil;
   end

   if (mode) then
      drag.mode = mode;
      -- Forget any buttons left over
      drag.LeftButton = nil;
      drag.RightButton = nil;
      -- Button count
      drag.buttons = 1;
      drag[arg1] = 1;
      drag:Show();
      drag.ox, drag.oy = ox, oy;
      drag.parent = this;
   end
   
   if (needUpdate) then
      StatRingsOptionsDrag_UpdateFrame(this);
   end
end

function StatRingsOptionsDrag_OnMouseUp()
   local drag = this.drag;
   if (not drag.mode) then
      return;
   end
   -- Check button is thought to be down
   if (not drag[arg1]) then
      return;
   end
   drag[arg1] = nil;
   drag.buttons = drag.buttons - 1;
   if (drag.buttons == 0) then
      StatRingsOptionsDrag_Drag_OnUpdate(drag);
      drag.mode = nil;
      drag:Hide();
   end
end

function StatRingsOptionsDrag_Drag_OnUpdate(this)
   local ox,oy = GetCursorPosition();
   if ((ox == this.ox) and (oy == this.oy)) then
      return;
   end
   this.ox,this.oy = ox,oy;

   local anchor = this.parent.anchor;
   local mode = this.mode;
   local dx,dy = ox-this.startx, oy-this.starty;
   local snapTo = not IsShiftKeyDown();
   
   if (mode == "move") then
      local oscale = this.ofsScale;
      dx,dy = dx / oscale, dy / oscale;
      local nro = sqrt(dx*dx+dy*dy) / STANDARD_RADIUS;
      local na;
      SetEffectiveLength(anchor, nro);
      if (snapTo) then
	 anchor.length = DoSnapTo(anchor.length, OFFSET_SNAPTO);
      end
      if (anchor.length > 0) then
	 na = deg(atan2(dx,dy));
	 if (na < 0) then
	    na = 360 + na;
	 end
      else
	 na = 0;
      end
      if (snapTo) then
	 na = DoSnapTo(na, ANGLE_SNAPTO);
      end
      if (anchor.relSpin and (not IsAltKeyDown())) then
	 local nSpin = mod(anchor.spin + anchor.angle - na, 360);
	 if (nSpin < 0) then nSpin = nSpin + 360; end
	 anchor.spin = nSpin;
      end
      anchor.angle = na;
      StatRingsOptionsDrag_UpdateFrame(this.parent);
      return;
   end
   if (mode == "scaleRotate") then
      local nlen = sqrt(dx*dx+dy*dy);

      if (this.doScale) then
	 local nscale = nlen / this.baseLen;
	 if (snapTo) then
	    nscale = DoSnapTo(nscale, SCALE_SNAPTO);
	 end
	 if (nscale < MIN_SCALE) then
	    nscale = MIN_SCALE;
	 end

	 -- Recalculate the length
	 local ro = GetEffectiveLength(anchor) * anchor.scale;
	 anchor.scale = nscale;
	 SetEffectiveLength(anchor, ro / nscale);
      end
      
      if ((nlen > 0) and (this.doAngle)) then
	 local na = deg(atan2(dx,dy));
	 na = na + this.ofsAngle;
	 if (snapTo) then
	    na = DoSnapTo(na, ANGLE_SNAPTO);
	 end
	 while (na < 360) do na = na + 360; end
	 na = math.mod(na, 360);
	 anchor.spin = na;
      end
      StatRingsOptionsDrag_UpdateFrame(this.parent);
      return;
   end
end

local ENABLED_PREVIEW = {
   shown = true,
   healthPct = 1,
   manaPct = 1,
   manaMax = 100,
   alpha = 1,
   unitName = true, -- Special value to show unit ID
};

local DISABLED_PREVIEW = {
   shown = true,
   force = "grey",
   alpha = 0.6,
   isGrey = true, -- Force show in greyscale
};

local function StatRingsOptionsDrag_Cmd(msg)
   local scaffold = StatRingsCluster.scaffold;

   if (msg == nil or msg == "") then
      StatRingsOptionsDragFrame.drag.mode = nil;
      StatRingsOptionsDragFrame.drag:Hide();
      StatRingsOptionsDragFrame:Hide();
      DEBUG("Cancelling any pending drag.");
      StatRingsCluster:RestoreInfoModel();
      return;
   end
   
   local anchorName = msg;
   local anchor = scaffold:GetAnchor(anchorName);
   if (not anchor) then
      DEBUG("Unable to find anchor '" .. anchorName .. "'");
      return;
   end
   
   local parentName = scaffold:GetParent(anchorName);
   local x,y,scale,spin = scaffold:GetLocation(parentName);
   if (not x) then
      DEBUG("Anchor parent is not visible");
      return;
   end

   StatRingsCluster:PreviewInfoModel(function() return true; end,
				     ENABLED_PREVIEW,
				     DISABLED_PREVIEW);

   StatRingsOptionsDrag_Start("Adjusting " .. anchorName,
			      StatRingsCluster,
			      x, y, scale / 100, spin,
			      anchor,
			      function(newAnchor) 
				 scaffold:SetAnchor(anchorName, newAnchor);
				 scaffold:Update();
			      end);
end

-- SlashCmdList["STATRINGS_OPTIONSDRAG"] = StatRingsOptionsDrag_Cmd;
-- SLASH_STATRINGS_OPTIONSDRAG1 = "/srdrag";
